# -*- coding: utf-8 -*-
"""
Created on Fri Jan 18 14:19:27 2019

@author: Guest Group
"""
from .free_property import FreeProperty

__all__ = ('Item', 'Attribute')


# %% Simple Property Classes                

class Attribute(FreeProperty):
    """Create an Attribute class that acts as a proxy for the value of an attribute.
    
    **Parameters**
    
        **obj:** object the attribute belongs to.
        
        **key:** attribute key.
        
    **Examples**
    
        First we need an object with attributes. So lets create a dummy class for this and initialize it:
        
        .. code-block:: python
        
            >>> class Dummy:
            ...    
            ...     def __init__(self, value):
            ...         self.value = value
            ...     
            ...     def __str__(self):
            ...         return 'dummy'
            ...
            >>> dummy = Dummy(20)
            
        Create an Attribute object with the object and the key:
            
        .. code-block:: python
    
            >>> v = Attribute(dummy, 'value')
            >>> v
            dummy.value -> 20
            
        The attribute object acts as a proxy for getting and setting the value:
        
        .. code-block
            
            >>> dummy.value = 0
            >>> v
            dummy.value -> 0
            >>> v += 40
            >>> dummy.value
            40
    
    """
    
    __slots__ = ('obj', 'key')
    
    def __init__(self, obj, key):
        self.obj = obj
        self.key = key
    
    @property
    def value(self):
        return getattr(self.obj, self.key)
    
    @value.setter
    def value(self, value):
        setattr(self.obj, self.key, value)
    
    def __repr__(self):
        return f'{self.obj}.{self.key} -> {self.value}'
    

class Item(FreeProperty):
    """Create an Item class that acts as a proxy for the value of an item.
    
    **Parameters**
    
        **obj:** object the item belongs to.
        
        **key:** item key.
        
    **Examples**
    
        First we need an object with items. So let's create a dictionary with items:
            
        .. code-block:: python
        
            >>> dummy = {'value': 20}
        
        Create an Item object with the object and the key:
            
        .. code-block:: python
    
            >>> v = Item(dummy, 'value')
            >>> v
            {'value': 20}['value'] -> 20
            
        The attribute object acts as a proxy for getting and setting the value:
        
        .. code-block
            
            >>> dummy['value'] = 0
            >>> v
            {'value': 20}['value'] -> 0
            >>> v += 40
            >>> {'value': 20}['value']
            40
    
    """
    __slots__ = ('obj', 'key')
    
    def __init__(self, obj, key):
        self.obj = obj
        self.key = key
    
    @property
    def value(self):
        return self.obj.__getitem__(self.key)
    
    @value.setter
    def value(self, value):
        self.obj.__setitem__(self.key, value)
    
    def __repr__(self):
        key = self.key
        if isinstance(key, str): key = f"'{key}'"
        return f'{self.obj}[{key}] -> {self.value}'
    
    
    


# -*- coding: utf-8 -*-
"""
Created on Fri Jan 18 14:18:27 2019

@author: Guest Group
"""

from .free_property import metaProperty, FreeProperty

__all__ = ('PropertyFactory',)

# %% Functions

def search_units(doc):
    # Find units
    par1 = doc.find('(')
    par2 = doc.find(')')
    if par1==-1 or par2==-1:
        units = ''
    else:
        units = doc[par1+1:par2]
    
    return units


# %% Property Factory


def PropertyFactory(fget, fset=None, clsname=None, doc=None, units=None):
    """Create an FreeProperty subclass with getter and setter functions.
    
    **Parameters**
    
        **fget:** [function] Should return value of instances.
        
        **fset:** [function] Should set the value of instances.
        
        **clsname:** [str] Name of the class. If None, the function name of fget will be used
        
        **doc:** [str] Docstring of class. If None, the docstring of fget will be used.
        
        **units:** [str] Units of measure. If None, the docstring will be searched and units of measure in parenthesis will be used.
    
    **Examples**
    
        The PropertyFactory is an FreeProperty class creator that functions just like Python 'property' objects. Use the PropertyFactory to create a Weight class which calculates weight based on density and volume:
        
        .. code-block:: python
            
            >>> def getter(self):
            ...    '''Weight (kg) based on volume (m^3).'''
            ...    data = self.data
            ...    rho = data['rho'] # Density (kg/m^3)
            ...    vol = data['vol'] # Volume (m^3)
            ...    return rho * vol
            >>>
            >>> def setter(self, weight):
            ...    data = self.data
            ...    rho = data['rho'] # Density (kg/m^3)
            ...    data['vol'] = weight / rho
            >>>
            >>> # Initialize with a value getter, setter, and the class name.
            >>> Weight = PropertyFactory(getter, setter, 'Weight')
            
        It is more convinient to use the PropertyFactory as a decorator:
        
        .. code-block:: python
           
            >>> @PropertyFactory
            >>> def Weight(self):
            ...    '''Weight (kg) based on volume (m^3).'''
            ...    data = self.data
            ...    rho = data['rho'] # Density (kg/m^3)
            ...    vol = data['vol'] # Volume (m^3)
            ...    return rho * vol
            >>>
            >>> @Weight.setter
            >>> def Weight(self, weight):
            ...    data = self.data
            ...    rho = data['rho'] # Density (kg/m^3)
            ...    data['vol'] = weight / rho
           
        
        Create dictionaries of data and initialize new Weight objects:
       
        .. code-block:: python
       
           >>> water_data = {'rho': 1000, 'vol': 3}
           >>> ethanol_data = {'rho': 789, 'vol': 3}
           >>> weight_water = Weight('Water', water_data)
           >>> weight_ethanol = Weight('Ethanol', ethanol_data)
           >>> weight_water
           Weight(Water) -> 3000 (kg)
           >>>weight_ethanol
           Weight(Ethanol) -> 2367 (kg)
        
        These properties behave just like their dynamic value:
        
        .. code-block:: python
        
            >>> weight_water + 30
            3030
            >>> weight_water + weight_ethanol
            5367
            >>> weight_water.bit_length() # A method of the dynamic value
            12
            
        Get and set the value through the 'value' attribute:
            
        .. code-block:: python
        
            >>> weight_water.value
            3000
            >>> weight_water.value = 4000 
            >>> weight_water.value
            4000
            >>> water_data # Note that the volume changed too
            {'rho': 1000, 'vol': 4}

        In place magic methods will also change the property value:
        
        .. code-block:: python
        
            >>> weight_water -= 1000
            >>> weight_water
            Weight(Water) -> 3000 (kg)
            >>> water_data  # The change also affects the original data
            {'rho': 1000, 'vol': 3}
            
    """       
    # Defaults
    if clsname is None: clsname = fget.__name__
    if doc     is None: doc     = fget.__doc__
    if units   is None: units   = search_units(doc)
    
    definitions = {'__doc__': doc,
                   '__slots__': ('name', 'data'),
                   '_units': units,
                   'value': property(fget, fset)}    
    
    return metaProperty(clsname, (FreeProperty,), definitions)



# -*- coding: utf-8 -*-
"""
Created on Fri Jan 18 14:20:29 2019

@author: Guest Group
"""

__all__ = ('FreeProperty' , 'Item', 'Attribute', 'PropertyFactory')

from .free_property import FreeProperty
from .subclasses import Item, Attribute
from .property_factory import PropertyFactory